#define com_maroontress_libcgi_Request_IMPORT
#define com_maroontress_libcgi_mime_ContentType_IMPORT
#define com_maroontress_libcgi_mime_Parameter_IMPORT
#define com_maroontress_libcgi_request_FileSegment_IMPORT
#define com_maroontress_libcgi_request_MimePart_IMPORT
#define com_maroontress_libcgi_request_MultipartPostRequest_IMPORT
#define std_io_OutputTube_IMPORT
#define std_lang_String_IMPORT
#define std_util_ArraySet_IMPORT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <com/maroontress/libcgi/Request.h>
#include <com/maroontress/libcgi/mime/ContentType.h>
#include <com/maroontress/libcgi/mime/Parameter.h>
#include <com/maroontress/libcgi/request/FileSegment.h>
#include <com/maroontress/libcgi/request/MultipartPostRequest.h>
#include <com/maroontress/libcgi/request/MimePart.h>
#include <std/io/OutputTube.h>
#include <std/lang/String.h>
#include <std/util/ArraySet.h>

struct MultipartPostRequest {
    struct Request super;
    union ArraySet partSet;
    FILE *fp;
};

union Request_Impl {
    struct Request request;
    struct MultipartPostRequest multipartPostRequest;
};

static void
Request_Impl_destroy(union Request_Impl *impl)
{
    struct MultipartPostRequest *r = &impl->multipartPostRequest;
    unsigned int k, n;

    if (r->fp != NULL) {
	fclose(r->fp);
    }
    if (r->partSet.impl != NULL) {
	n = ArraySet_getSize(r->partSet);
	for (k = 0; k < n; ++k) {
	    MimePart_free((struct MimePart *)ArraySet_get(r->partSet, k));
	}
	ArraySet_free(r->partSet);
    }
    free(r);
}

static struct MimePart *
getMimePartByString(struct MultipartPostRequest *r, struct String *key)
{
    size_t k, n;
    struct ContentType *ct;
    struct Parameter *param;
    struct MimePart *mp;

    n = ArraySet_getSize(r->partSet);
    for (k = 0; k < n; ++k) {
	mp = (struct MimePart *)ArraySet_get(r->partSet, k);
	if ((ct = MimePart_getDisposition(mp)) == NULL
	    || (param = ContentType_getParameter(ct, "name")) == NULL
	    || Parameter_getValue(param) != key) {
	    continue;
	}
	return mp;
    }
    return NULL;
}

static struct MimePart *
getMimePart(struct MultipartPostRequest *r, const char *key)
{
    struct String *canonicalKey;
    struct MimePart *mp;

    if ((canonicalKey = String_intern(key)) == NULL)
	return NULL;
    mp = getMimePartByString(r, canonicalKey);
    String_free(canonicalKey);
    return mp;
}

static char *
Request_Impl_getValue(union Request_Impl *impl, const char *key)
{
    struct MultipartPostRequest *r = &impl->multipartPostRequest;
    struct MimePart *mp;

    if ((mp = getMimePart(r, key)) == NULL)
	return NULL;
    return MimePart_getContent(mp);
}

static struct ContentType *
Request_Impl_getContentType(union Request_Impl *impl, const char *key)
{
    struct MultipartPostRequest *r = &impl->multipartPostRequest;
    struct MimePart *mp;

    if ((mp = getMimePart(r, key)) == NULL)
	return NULL;
    return MimePart_getContentType(mp);
}

static int
Request_Impl_writeValue(union Request_Impl *impl, const char *key,
			union OutputTube out)
{
    struct MultipartPostRequest *r = &impl->multipartPostRequest;
    struct MimePart *mp;

    if ((mp = getMimePart(r, key)) == NULL)
	return -1;
    return (MimePart_write(mp, out) < 0) ? -1 : 0;
}

static const struct Request_Methods METHODS = {
    Request_Impl_destroy,
    Request_Impl_getValue,
    Request_Impl_getContentType,
    Request_Impl_writeValue};

static int
saveFile(struct MultipartPostRequest *r)
{
    char template[] = "/tmp/com_maroontress_libcgi_XXXXXX";
    char data[4096];
    size_t size;
    int fd;
    
    if ((fd = mkstemp(template)) < 0)
        return -1;
    if (unlink(template) < 0 || (r->fp = fdopen(fd, "w+")) == NULL) {
        close(fd);
        return -1;
    }
    while ((size = fread(data, 1, sizeof(data), stdin)) > 0
	   && fwrite(data, size, 1, r->fp) == 1) {
	continue;
    }
    return (ferror(stdin) || ferror(r->fp) || fflush(r->fp)) ? -1 : 0;
}

static int
seekToFirstBoundary(FILE *fp, const char *top, size_t len,
		    struct FileSegment *seg)
{
    int c;
    const char *b;

    if ((seg->offset = ftello(fp)) < 0)
	return -1;
    for (;;) {
        for (b = top; *b != 0 && (c = getc(fp)) != EOF && *b == c; ++b)
            ;
        if (*b == 0)
            break;
        if (c == EOF)
            return -1;
        while (c != '\n' && (c = getc(fp)) != EOF)
            ;
        if (c == EOF)
            return -1;
    }
    return ((seg->length = ftello(fp) - seg->offset - len) < 0) ? -1 : 0;
}

static int
seekToNextBoundary(FILE *fp, const char *top, size_t len,
		   struct FileSegment *seg)
{
    int c;
    const char *b;

    for (b = "--"; *b != 0 && (c = getc(fp)) != EOF && *b == c; ++b)
        ;
    if (*b == 0)
        return 1;
    if (c == EOF)
        return -1;
    while (c != '\n' && (c = getc(fp)) != EOF)
        ;
    if (c == EOF
	|| seekToFirstBoundary(fp, top, len, seg) < 0
	|| seg->length < 2) {
        return -1;
    }
    /*
      seg->length϶ľCRLFΥޤࡣ

      seekToFirstBoundary()Ƥ:
      | --boundary CR LF
      | 

      seekToFirstBoundary()Ƥ:
      | --boundary CR LF
      | data CR LF
      | --boundary
    */
    seg->length -= 2;
    return 0;
}

static int
loadAllParts(struct MultipartPostRequest *r, const char *boundary, size_t len)
{
    struct FileSegment seg;
    struct MimePart *mp;
    off_t offset;
    int s;

    rewind(r->fp);
    if (seekToFirstBoundary(r->fp, boundary, len, &seg) != 0) {
        return -1;
    }
    while ((s = seekToNextBoundary(r->fp, boundary, len, &seg)) == 0) {
	if ((offset = ftello(r->fp)) < 0)
	    return -1;
        if ((mp = MimePart_create(r->fp, &seg)) == NULL) {
            return -1;
        }
	if (ArraySet_add(r->partSet, mp) < 0) {
	    MimePart_free(mp);
            return -1;
	}
	if (fseeko(r->fp, offset, SEEK_SET) < 0)
	    return -1;
    }
    return (s < 0) ? -1 : 0;
}

static int
initializeWithBoundary(struct MultipartPostRequest *r,
		       const char *boundary, size_t len)
{
    if ((r->partSet = ArraySet_create()).impl == NULL
	|| saveFile(r) < 0
	|| loadAllParts(r, boundary, len) < 0) {
	return -1;
    }
    return 0;
}

static int
initialize(struct MultipartPostRequest *r, const char *boundaryId)
{
    size_t size;
    char *boundary;
    int s;

    size = strlen(boundaryId) + 3;
    if ((boundary = malloc(size)) == NULL) {
	return -1;
    }
    snprintf(boundary, size, "--%s", boundaryId);
    s = initializeWithBoundary(r, boundary, size - 1);
    free(boundary);
    return 0;
}

struct Request *
MultipartPostRequest_create(const char *boundary)
{
    union Request_Impl *impl;

    if ((impl = (union Request_Impl *)malloc(sizeof(*impl))) == NULL)
	return NULL;
    Request_initialize(&impl->request, &METHODS, impl);
    impl->multipartPostRequest.partSet.impl = NULL;
    impl->multipartPostRequest.fp = NULL;

    if (initialize(&impl->multipartPostRequest, boundary) < 0) {
	Request_Impl_destroy(impl);
	return NULL;
    }
    return &impl->request;
}
